import os
import re

from eth_utils import (
    add_0x_prefix,
    remove_0x_prefix,
    is_integer,
)

from .filesystem import (
    normpath,
)
from .six import (
    parse,
)


BASE_BLOCKCHAIN_STORAGE_DIR = "./chains"


@normpath
def get_base_blockchain_storage_dir(project_dir):
    base_blochcain_storage_dir = os.path.join(project_dir, BASE_BLOCKCHAIN_STORAGE_DIR)
    return base_blochcain_storage_dir


BLOCK_OR_TRANSACTION_HASH_REGEX = "^(?:0x)?[a-zA-Z0-9]{64}$"


def is_block_or_transaction_hash(value):
    return bool(re.match(BLOCK_OR_TRANSACTION_HASH_REGEX, value))


BLOCK = 'block'
TRANSACTION = 'transaction'


def create_BIP122_uri(chain_id, resource_type, resource_identifier):
    """
    See: https://github.com/bitcoin/bips/blob/master/bip-0122.mediawiki
    """
    if resource_type not in {BLOCK, TRANSACTION}:
        raise ValueError("Invalid resource_type.  Must be one of 'block' or 'transaction'")
    elif not is_block_or_transaction_hash(resource_identifier):
        raise ValueError("Invalid resource_identifier.  Must be a hex encoded 32 byte value")
    elif not is_block_or_transaction_hash(chain_id):
        raise ValueError("Invalid chain_id.  Must be a hex encoded 32 byte value")

    return parse.urlunsplit([
        'blockchain',
        remove_0x_prefix(chain_id),
        "{0}/{1}".format(resource_type, remove_0x_prefix(resource_identifier)),
        '',
        '',
    ])


def create_block_uri(chain_id, block_identifier):
    if is_integer(block_identifier):
        return create_BIP122_uri(chain_id, 'block', str(block_identifier))
    else:
        return create_BIP122_uri(chain_id, 'block', remove_0x_prefix(block_identifier))


def create_transaction_uri(chain_id, transaction_hash):
    return create_BIP122_uri(chain_id, 'transaction', transaction_hash)


def get_chain_id(web3):
    return web3.eth.getBlock(0)['hash']


def get_chain_definition(web3, min_block_number=0, num_confirmations=0):
    """
    Return the blockchain URI that
    """
    chain_id = get_chain_id(web3)
    latest_block = web3.eth.getBlock('latest')
    latest_block_number = latest_block['number']

    target_block_number = latest_block_number - num_confirmations
    if target_block_number < min_block_number:
        raise ValueError("Cannot generate chain definition matching given constraints")

    block_for_definition = web3.eth.getBlock(target_block_number)
    block_hash_for_definition = block_for_definition['hash']

    return create_block_uri(chain_id, block_hash_for_definition)


BIP122_URL_REGEX = (
    "^"
    "blockchain://"
    "(?P<chain_id>[a-zA-Z0-9]{64})"
    "/"
    "(?P<resource_type>block|transaction)"
    "/"
    "(?P<resource_hash>[a-zA-Z0-9]{64})"
    "$"
)


def is_BIP122_uri(value):
    return bool(re.match(BIP122_URL_REGEX, value))


def parse_BIP122_uri(blockchain_uri):
    match = re.match(BIP122_URL_REGEX, blockchain_uri)
    if match is None:
        raise ValueError("Invalid URI format: '{0}'".format(blockchain_uri))
    chain_id, resource_type, resource_hash = match.groups()
    return (
        add_0x_prefix(chain_id),
        resource_type,
        add_0x_prefix(resource_hash),
    )


def is_BIP122_block_uri(value):
    if not is_BIP122_uri(value):
        return False
    _, resource_type, _ = parse_BIP122_uri(value)
    return resource_type == BLOCK


def is_BIP122_transaction_uri(value):
    if not is_BIP122_uri(value):
        return False
    _, resource_type, _ = parse_BIP122_uri(value)
    return resource_type == TRANSACTION


def check_if_chain_matches_chain_uri(web3, blockchain_uri):
    chain_id, resource_type, resource_hash = parse_BIP122_uri(blockchain_uri)
    genesis_block = web3.eth.getBlock('earliest')
    if genesis_block['hash'] != chain_id:
        return False

    if resource_type == BLOCK:
        resource = web3.eth.getBlock(resource_hash)
    elif resource_type == TRANSACTION:
        resource = web3.eth.getTransaction(resource_hash)
    else:
        raise ValueError("Unsupported resource type: {0}".format(resource_type))

    if resource['hash'] == resource_hash:
        return True
    else:
        return False


def is_synced(web3, allowed_block_delta=3):
    sync_info = web3.eth.syncing
    if not sync_info:
        return False
    blocks_to_be_synced = sync_info['highestBlock'] - sync_info['currentBlock']
    return blocks_to_be_synced <= allowed_block_delta



from populus.utils.empty import empty
from populus.utils.wait import (
    wait_for_block_number,
    wait_for_peers,
    wait_for_syncing,
    wait_for_transaction_receipt,
    wait_for_unlock,
)


class Wait(object):
    web3 = None
    timeout = 120
    poll_interval = None

    def __init__(self, web3, timeout=empty, poll_interval=empty):
        self.web3 = web3
        if timeout is not empty:
            self.timeout = timeout
        if poll_interval is not empty:
            self.poll_interval = poll_interval

    def for_contract_address(self, txn_hash, timeout=empty, poll_interval=empty):
        kwargs = {}
        if timeout is not empty:
            kwargs['timeout'] = timeout
        if poll_interval is not empty:
            kwargs['poll_interval'] = poll_interval

        kwargs.setdefault('timeout', self.timeout)
        kwargs.setdefault('poll_interval', self.poll_interval)

        txn_receipt = self.for_receipt(txn_hash, **kwargs)
        return txn_receipt['contractAddress']

    def for_receipt(self, txn_hash, timeout=empty, poll_interval=empty):
        kwargs = {}

        if timeout is not empty:
            kwargs['timeout'] = timeout
        if poll_interval is not empty:
            kwargs['poll_interval'] = poll_interval

        kwargs.setdefault('timeout', self.timeout)
        kwargs.setdefault('poll_interval', self.poll_interval)

        return wait_for_transaction_receipt(self.web3, txn_hash, **kwargs)

    def for_block(self, block_number=empty, timeout=empty, poll_interval=empty):
        kwargs = {}

        if block_number is not empty:
            kwargs['block_number'] = block_number
        if timeout is not empty:
            kwargs['timeout'] = timeout
        if poll_interval is not empty:
            kwargs['poll_interval'] = poll_interval

        kwargs.setdefault('timeout', self.timeout)
        kwargs.setdefault('poll_interval', self.poll_interval)

        return wait_for_block_number(self.web3, **kwargs)

    def for_unlock(self, account=empty, timeout=empty, poll_interval=empty):
        kwargs = {}

        if account is not empty:
            kwargs['account'] = account
        if timeout is not empty:
            kwargs['timeout'] = timeout
        if poll_interval is not empty:
            kwargs['poll_interval'] = poll_interval

        kwargs.setdefault('timeout', self.timeout)
        kwargs.setdefault('poll_interval', self.poll_interval)

        return wait_for_unlock(self.web3, **kwargs)

    def for_peers(self, peer_count=empty, timeout=empty, poll_interval=empty):
        kwargs = {}

        if peer_count is not empty:
            kwargs['peer_count'] = peer_count
        if timeout is not empty:
            kwargs['timeout'] = timeout
        if poll_interval is not empty:
            kwargs['poll_interval'] = poll_interval

        kwargs.setdefault('timeout', self.timeout)
        kwargs.setdefault('poll_interval', self.poll_interval)

        return wait_for_peers(self.web3, **kwargs)

    def for_syncing(self, timeout=empty, poll_interval=empty):
        kwargs = {}

        if timeout is not empty:
            kwargs['timeout'] = timeout
        if poll_interval is not empty:
            kwargs['poll_interval'] = poll_interval

        kwargs.setdefault('timeout', self.timeout)
        kwargs.setdefault('poll_interval', self.poll_interval)

        return wait_for_syncing(self.web3, **kwargs)



import sys

from watchdog.utils import UnsupportedLibc
from watchdog.observers.polling import PollingObserver
from watchdog.events import FileSystemEventHandler


def get_observer_for_platform(platform):
    if platform == 'darwin':
        from watchdog.observers.kqueue import KqueueObserver
        return KqueueObserver
    elif platform.startswith('linux'):
        try:
            from watchdog.observers.inotify import (
                InotifyObserver,
            )
        except UnsupportedLibc:
            return PollingObserver
        else:
            return InotifyObserver
    elif platform in {'win32', 'cygwin'}:
        from watchdog.observers.read_directory_changes import WindowsAPIObserver
        return WindowsAPIObserver
    else:
        raise ValueError(
            "Unsupported platfom '{0}'.  Must be one of 'win32', 'cgwin', "
            "'linux*', 'darwin'"
        )


class EventHandler(FileSystemEventHandler):
    def __init__(self, callback):
        self.callback = callback

    def on_any_event(self, event):
        self.callback(event.src_path, event.event_type)


class DirWatcher(object):
    def __init__(self, path_to_watch, callback, observer_class=None):
        if observer_class is None:
            observer_class = get_observer_for_platform(sys.platform)
        self.observer = observer_class()
        self.event_handler = EventHandler(callback)
        self.observer.schedule(self.event_handler, path_to_watch, recursive=True)

    def start(self):
        self.observer.start()

    def stop(self):
        self.observer.stop()



import os
import shutil

from populus.utils.filesystem import (
    ensure_path_exists,
)

from geth.wrapper import (
    construct_test_chain_kwargs,
)

from geth.accounts import (
    create_new_account,
)

from geth.install import (
    chmod_plus_x,
)


INIT_COMMAND = (
    "geth --rpc --rpcaddr 127.0.0.1 --rpcport 8545 "
    "--rpcapi admin,debug,eth,miner,net,personal,shh,txpool,web3,ws "
    "--ws --wsaddr 127.0.0.1 --wsport 8546 "
    "--wsapi admin,debug,eth,miner,net,personal,shh,txpool,web3,ws "
    "--datadir {data_dir} --maxpeers 0 --networkid 1234 --port 30303 "
    "--ipcpath {ipc_path} --nodiscover --mine --minerthreads 1 "
    "init {genesis_path}"
)

RUN_COMMAND = (
    "geth --rpc --rpcaddr 127.0.0.1 --rpcport 8545 "
    "--rpcapi admin,debug,eth,miner,net,personal,shh,txpool,web3,ws "
    "--ws --wsaddr 127.0.0.1 --wsport 8546 "
    "--wsapi admin,debug,eth,miner,net,personal,shh,txpool,web3,ws "
    "--datadir {data_dir} --maxpeers 0 --networkid 1234 --port 30303 "
    "--ipcpath {ipc_path} --unlock {account} "
    "--password {password_path} --nodiscover --mine --minerthreads 1 "
)

GENESIS_BLOCK = '''
{
    "parentHash": "0x0000000000000000000000000000000000000000000000000000000000000000",
  "coinbase": "%s",
   "extraData": "0x686f727365",
   "config": {
       "daoForkBlock": 0,
     "daoForSupport": true,
    "homesteadBlock": 0
    },
   "timestamp": "0x0",
    "mixhash": "0x0000000000000000000000000000000000000000000000000000000000000000",
    "nonce": "0xdeadbeefdeadbeef",
    "alloc": {
        "%s":{
          "balance": "1000000000000000000000000000000"
      }
      },
    "gasLimit": "0x47d5cc",
    "difficulty": "0x01"
}
'''


def new_local_chain(project_dir, chain_name):
    chains_path = os.path.join(project_dir, 'chains')
    ensure_path_exists(chains_path)
    chain_dir = os.path.join(chains_path, chain_name)
    data_dir = os.path.join(chain_dir, 'chain_data')
    overrides = {
        'data_dir': data_dir,
        'ws_port': '8546',
        'rpc_port': '8545',
        'port': '30303'
    }
    geth_kwargs = construct_test_chain_kwargs(**overrides)
    password = geth_kwargs.pop('password')
    data_dir = geth_kwargs.pop('data_dir')
    account = create_new_account(data_dir, password, **geth_kwargs)
    account = account.decode('ascii')

    genesis = GENESIS_BLOCK % (account, account,)
    genesis_path = os.path.join(chain_dir, 'genesis.json')
    with open(genesis_path, 'w+') as f:
        f.write(genesis)

    password_path = os.path.join(chain_dir, 'password')
    shutil.copyfile(password, password_path)

    ipc_path = geth_kwargs['ipc_path']
    init = INIT_COMMAND.format(
        data_dir=data_dir,
        ipc_path=ipc_path,
        genesis_path=genesis_path
    )
    init_geth_path = os.path.join(chain_dir, 'init_chain.sh')
    with open(init_geth_path, 'w+') as f:
        f.write(init)
    chmod_plus_x(init_geth_path)

    run = RUN_COMMAND.format(
        data_dir=data_dir,
        ipc_path=ipc_path,
        password_path=password_path,
        account=account
    )
    run_geth_path = os.path.join(chain_dir, 'run_chain.sh')
    with open(run_geth_path, 'w+') as f:
        f.write(run)
    chmod_plus_x(run_geth_path)



"""
Lifted from the base58 package.
"""
import sys


# 58 character alphabet used
ALPHABET = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'


if sys.version_info.major == 2:
    iseq = lambda s: map(ord, s)  # noqa: E731
    bseq = lambda s: ''.join(map(chr, s))  # noqa: E731
    buffer = lambda s: s  # noqa: E731
else:
    iseq = lambda s: s  # noqa: E731
    bseq = bytes  # noqa: E731
    buffer = lambda s: s.buffer  # noqa: E731


def b58encode(v):
    origlen = len(v)
    v = v.lstrip(b'\0')
    newlen = len(v)

    p, acc = 1, 0
    for c in iseq(v[::-1]):
        acc += p * c
        p = p << 8

    result = ''
    while acc > 0:
        acc, mod = divmod(acc, 58)
        result += ALPHABET[mod]

    return (result + ALPHABET[0] * (origlen - newlen))[::-1]


def b58decode(v):
    if not isinstance(v, str):
        v = v.decode('ascii')

    origlen = len(v)
    v = v.lstrip(ALPHABET[0])
    newlen = len(v)

    p, acc = 1, 0
    for c in v[::-1]:
        acc += p * ALPHABET.index(c)
        p *= 58

    result = []
    while acc > 0:
        acc, mod = divmod(acc, 256)
        result.append(mod)

    return (bseq(result) + b'\0' * (origlen - newlen))[::-1]



from populus.contracts.exceptions import (
    UnknownContract,
)
from populus.contracts.contract import (
    construct_contract_factory,
)


class BaseContractBackend(object):
    chain = None

    def __init__(self, chain, config):
        self.chain = chain
        self.config = config
        self.setup_backend()

    #
    # Meta API
    #
    @property
    def is_provider(self):
        raise NotImplementedError("Must be implemented by subclasses")

    @property
    def is_registrar(self):
        raise NotImplementedError("Must be implemented by subclasses")

    def setup_backend(self):
        """
        Hook for subclasses to do backend initialization without having to
        override the `__init__` method.
        """
        pass

    #
    # Registrar API
    #
    def set_contract_address(self, instance_name, address):
        """
        Returns the address for the contract instance in the registrar.
        """
        raise NotImplementedError("Must be implemented by subclasses")

    def get_contract_addresses(self, instance_name):
        """
        Returns all known address of the requested contract instance.
        """
        raise NotImplementedError("Must be implemented by subclasses")

    #
    # Provider API
    #
    def get_contract_identifier(self, contract_name):
        """
        Given a contract name, return the contract identifier that will
        retrieve the appropriate the base contract factory.
        """
        raise NotImplementedError("Must be implemented by subclasses")

    def get_base_contract_factory(self, contract_name):
        """
        Returns the base contract factory for the given contract identifier.
        """
        try:
            contract_identifier = self.get_contract_identifier(contract_name)
        except KeyError:
            raise UnknownContract(
                "No contract available for the name '{0}'".format(
                    contract_name,
                )
            )

        contract_data = self.get_contract_data(contract_identifier)

        base_contract_factory = construct_contract_factory(
            chain=self.chain,
            contract_identifier=contract_identifier,
            contract_data=contract_data,
        )
        return base_contract_factory

    def get_all_contract_data(self):
        """
        Returns all contract data available from this backend.
        """
        raise NotImplementedError("Must be implemented by subclasses")

    def get_contract_data(self, contract_name):
        """
        Returns the raw contract data.
        """
        contract_identifier = self.get_contract_identifier(contract_name)
        try:
            return self.get_all_contract_data()[contract_identifier]
        except KeyError:
            raise UnknownContract(
                "No contract found for the name '{0}'".format(contract_name)
            )

    def get_all_contract_names(self):
        """
        Returns a set of all of the contract names for this backend.
        """
        return set(self.get_all_contract_data().keys())



import re
import functools

from cytoolz.functoolz import (
    compose,
)

from eth_utils import (
    coerce_args_to_text,
    remove_0x_prefix,
    add_0x_prefix,
    to_tuple,
)

from .formatting import (
    remove_dunderscore_prefix,
)


DEPENDENCY_RE = (
    r'__'  # Prefixed by double underscore
    r'.{36}'  # 36 characters of dubious character
    r'__'  # End with a double underscore
)


# start and length should be `byte` offsets meaning they represent the
# start/length in the bytecode in its bytes representation.  To transate to hex
# representation, these two numbers should be multiplied by two.
def LinkReference(source_path, name, start, length):
    return {
        'source_path': source_path,
        'name': name,
        'start': start,
        'length': length,
    }


#
# Standard JSON utils
#
@to_tuple
def normalize_standard_json_link_references(link_references):
    for source_path, names in link_references.items():
        for contract_name, reference_locations in names.items():
            for location in reference_locations:
                yield LinkReference(
                    source_path=source_path,
                    name=contract_name,
                    start=location['start'] * 2,  # convert binary offsets to hex
                    length=location['length'] * 2,  # convert binary offsets to hex
                )


#
# Combined JSON utils
#
def remove_dunderscore_wrapper(value):
    return remove_dunderscore_prefix(value.rstrip('_'))


@to_tuple
@coerce_args_to_text
def find_placeholder_locations(bytecode):
    """
    Given bytecode, this will return all of the linked references from within
    the bytecode.
    """
    unprefixed_bytecode = remove_0x_prefix(bytecode)

    for match in re.finditer(DEPENDENCY_RE, unprefixed_bytecode):
        start = match.start()
        length = match.end() - start
        placeholder = unprefixed_bytecode[start:start + length]
        yield (remove_dunderscore_wrapper(placeholder), start, length)


def expand_placeholder(placeholder, full_names):
    """
    Link references whos names are longer than their bytecode representations
    will get truncated to 4 characters short of their full name because of the
    double underscore prefix and suffix.  This embedded string is referred to
    as the `placeholder`

    This expands `placeholder` to it's full reference name or raise a value
    error if it is unable to find an appropriate expansion.
    """
    if placeholder in full_names:
        return placeholder

    candidates = [
        full_name
        for full_name
        in full_names
        if full_name.startswith(placeholder)
    ]
    if len(candidates) == 1:
        return candidates[0]
    elif len(candidates) > 1:
        raise ValueError(
            "Multiple candidates found trying to expand '{0}'.  Found '{1}'. "
            "Searched '{2}'".format(
                placeholder,
                ", ".join(candidates),
                ", ".join(full_names),
            )
        )
    else:
        raise ValueError(
            "Unable to expand '{0}'. "
            "Searched {1}".format(
                placeholder,
                ", ".join(full_names),
            )
        )


@to_tuple
def normalize_placeholder_link_references(placeholder_locations, compiled_contracts):
    all_contract_names = set(
        contract_data['name']
        for contract_data
        in compiled_contracts
    )
    contract_source_paths = {
        contract_data['name']: contract_data['source_path']
        for contract_data
        in compiled_contracts
    }

    for placeholder, start, length in placeholder_locations:
        contract_name = expand_placeholder(placeholder, all_contract_names)
        source_path = contract_source_paths[contract_name]
        yield LinkReference(
            source_path=source_path,
            name=contract_name,
            start=start,
            length=length,
        )


#
# Bytecode Linking utils
#
def insert_link_value(bytecode, value, offset):
    return add_0x_prefix(''.join((
        remove_0x_prefix(bytecode)[:offset],
        remove_0x_prefix(value),
        remove_0x_prefix(bytecode)[offset + len(remove_0x_prefix(value)):]
    )))


def link_bytecode(bytecode, link_reference_values):
    """
    Given the bytecode for a contract, and it's dependencies in the form of
    {contract_name: address} this functino returns the bytecode with all of the
    link references replaced with the dependency addresses.

    TODO: validate that the provided values are of the appropriate length
    """
    linker_fn = compose(*(
        functools.partial(
            insert_link_value,
            value=value,
            offset=link_reference['start'],
        )
        for link_reference, value in link_reference_values
    ))
    linked_bytecode = linker_fn(bytecode)
    return linked_bytecode



from __future__ import absolute_import

from eth_utils import (
    is_string,
)

from populus.utils.module_loading import (
    import_string,
)
from populus.config.helpers import (
    ClassImportPath,
)

from .base import Config


BACKEND_IDENTIFIER_MAP = {
    'solc:combined-json': 'populus.compilation.backends.solc.SolcCombinedJSONBackend',
    'solc:standard-json': 'populus.compilation.backends.solc.SolcStandardJSONBackend',
}


UNSUPPORTED_BACKEND_IDENTIFIER_MSG = (
    "Unsupported type. Must be either a backend class, a dot "
    "separated python path to a backend class, or one of `ipc` or "
    "`rpc`"
)


class CompilerConfig(Config):
    backend_class = ClassImportPath('class')

    def set_backend_class(self, backend_identifier):
        if isinstance(backend_identifier, type):
            self.backend_class = backend_identifier
        elif is_string(backend_identifier):
            if backend_identifier.lower() in BACKEND_IDENTIFIER_MAP:
                self.backend_class = BACKEND_IDENTIFIER_MAP[backend_identifier.lower()]
            else:
                try:
                    import_string(backend_identifier)
                except ImportError:
                    raise ValueError(
                        UNSUPPORTED_BACKEND_IDENTIFIER_MSG.format(backend_identifier)
                    )
                else:
                    self.backend_class = backend_identifier
        else:
            raise ValueError(UNSUPPORTED_BACKEND_IDENTIFIER_MSG.format(backend_identifier))

    @property
    def backend(self):
        return self.backend_class(self.backend_settings)

    @property
    def backend_settings(self):
        return self.get('settings', {})



from __future__ import absolute_import

from populus.utils.module_loading import (
    import_string,
)
from populus.utils.types import (
    is_string,
)
from populus.config.helpers import (
    ClassImportPath,
)

from .base import Config


UNSUPPORTED_BACKEND_IDENTIFIER_MSG = (
    "Unsupported type. Must be either a backend class, a dot "
    "separated python path to a backend class, or one of {0}"
)


class BaseBackendConfig(Config):
    backend_class = ClassImportPath('class')
    backend_class_shortnames = None

    def set_backend_class(self, backend_identifier):
        if isinstance(backend_identifier, type):
            self.backend_class = backend_identifier
        elif is_string(backend_identifier):
            if backend_identifier.lower() in self.backend_class_shortnames:
                self.backend_class = self.backend_class_shortnames[backend_identifier.lower()]
            else:
                try:
                    import_string(backend_identifier)
                except ImportError:
                    raise ValueError(
                        UNSUPPORTED_BACKEND_IDENTIFIER_MSG.format(
                            ','.join(tuple(self.backend_class_shortnames.keys()))
                        )
                    )
                else:
                    self.backend_class = backend_identifier
        else:
            raise ValueError(
                UNSUPPORTED_BACKEND_IDENTIFIER_MSG.format(
                    ','.join(tuple(self.backend_class_shortnames.keys()))
                )
            )

    @property
    def backend(self):
        return self.backend_class(**self.backend_kwargs)

    @property
    def backend_kwargs(self):
        return self.get('settings', {})

    @backend_kwargs.setter
    def backend_kwargs(self, value):
        self['backend.settings'] = value

    @property
    def priority(self):
        return self['priority']

    @priority.setter
    def priority(self, value):
        self['priority'] = value


class ContractBackendConfig(BaseBackendConfig):
    backend_class_shortnames = {
        'memory': 'populus.contracts.backends.memory.MemoryBackend',
        'jsonfile': 'populus.contracts.backends.filesystem.JSONFileBackend',
        'project': 'populus.contracts.backends.project.ProjectContractsBackend',
    }



import copy

from populus.utils.networking import (
    wait_for_connection,
    get_open_port,
)

from .base import (
    BaseChain,
)


class TestRPCChain(BaseChain):
    rpc_port = None

    def get_web3_config(self):
        base_config = super(TestRPCChain, self).get_web3_config()
        config = copy.deepcopy(base_config)
        config['provider.settings.port'] = self.rpc_port
        return config

    def __enter__(self):
        if self._running:
            raise ValueError("The TesterChain is already running")

        if self.rpc_port is None:
            self.rpc_port = get_open_port()

        self._running = True

        self.rpc_methods = self.web3.providers[0].server.application.rpc_methods

        self.rpc_methods.full_reset()
        self.rpc_methods.rpc_configure('eth_mining', False)
        self.rpc_methods.rpc_configure('eth_protocolVersion', '0x3f')
        self.rpc_methods.rpc_configure('net_version', 1)
        self.rpc_methods.evm_mine()

        wait_for_connection('127.0.0.1', self.rpc_port)
        return self

    def __exit__(self, *exc_info):
        if not self._running:
            raise ValueError("The TesterChain is not running")
        try:
            self.web3.providers[0].server.stop()
            self.web3.providers[0].server.close()
            self.web3.providers[0].thread.kill()
        except AttributeError:
            self.web3.providers[0].server.shutdown()
            self.web3.providers[0].server.server_close()
        finally:
            self._running = False



import collections
import functools
import itertools

from .string import (
    normalize_class_name,
)


class cached_property(object):
    """
    Source: Django (django.utils.functional)

    Decorator that converts a method with a single self argument into a
    property cached on the instance.

    Optional ``name`` argument allows you to make cached properties of other
    methods. (e.g.  url = cached_property(get_absolute_url, name='url') )
    """
    def __init__(self, func, name=None):
        self.func = func
        self.__doc__ = getattr(func, '__doc__')
        self.name = name or func.__name__

    def __get__(self, instance, type=None):
        if instance is None:
            return self
        res = instance.__dict__[self.name] = self.func(instance)
        return res


def noop(*args, **kwargs):
    pass


def to_object(class_name, bases=None):
    if bases is None:
        bases = (object,)

    def outer(fn):
        @functools.wraps(fn)
        def inner(*args, **kwargs):
            props = fn(*args, **kwargs)
            return type(normalize_class_name(class_name), bases, props)
        return inner
    return outer


def apply_to_return_value(callback):
    def outer(fn):
        @functools.wraps(fn)
        def inner(*args, **kwargs):
            return callback(fn(*args, **kwargs))

        return inner
    return outer


chain_return = apply_to_return_value(itertools.chain.from_iterable)
to_set = apply_to_return_value(set)


def get_duplicates(values):
    duplicates = {
        key
        for key, value
        in collections.Counter(values).items()
        if value > 1
    }
    return duplicates



from __future__ import absolute_import

import copy

from populus.config.defaults import (
    load_default_config,
)

from populus.config.validation import (
    get_validation_errors,
    format_errors,
)

from populus.config.versions import (
    V6,
    V7,
)

from populus.config import (
    Config,
)


def upgrade_v6_to_v7(v6_config):
    """
    Upgrade a v6 config file to a v7 config file.
    """
    errors = get_validation_errors(v6_config, version=V6)
    if errors:
        raise ValueError(
            "Cannot upgrade invalid config.  Please ensure that your current "
            "configuration file is valid:\n\n{0}".format(
                format_errors(errors),
            )
        )

    v6_default = load_default_config(version=V6)
    v7_default = load_default_config(version=V7)

    v6_default_config = Config(v6_default)
    v6_default_config.unref()

    v7_default_config = Config(v7_default)
    v7_default_config.unref()

    if v6_config == v6_default_config:
        return v7_default_config

    # V7 just moved to user config, no change in keys
    upgraded_v6_config = copy.deepcopy(v6_config)
    upgraded_v6_config['version'] = V7

    return upgraded_v6_config



import pprint
import os

from .base import (
    BaseCompilerBackend,
)


class ViperBackend(BaseCompilerBackend):
    project_source_glob = ('*.v.py', '*.vy')
    test_source_glob = ('test_*.v.py', 'test_*.vy')

    def get_compiled_contracts(self, source_file_paths, import_remappings):
        try:
            from viper import compiler
        except ImportError:
            raise ImportError(
                'viper needs to be installed to use ViperBackend' +
                ' as compiler backend.'
            )

        self.logger.debug("Compiler Settings: %s", pprint.pformat(self.compiler_settings))

        compiled_contracts = []

        for contract_path in source_file_paths:
            code = open(contract_path).read()
            abi = compiler.mk_full_signature(code)
            bytecode = '0x' + compiler.compile(code).hex()
            bytecode_runtime = '0x' + compiler.compile(code, bytecode_runtime=True).hex()
            compiled_contracts.append({
                'name': os.path.basename(contract_path).split('.')[0],
                'abi': abi,
                'bytecode': bytecode,
                'bytecode_runtime': bytecode_runtime,
                'linkrefs': [],
                'linkrefs_runtime': [],
                'source_path': contract_path
            })

        return compiled_contracts



from __future__ import absolute_import

import copy
import pprint

from populus.utils.mappings import (
    set_nested_key,
    get_nested_key,
    has_nested_key,
)

from populus.config.defaults import (
    load_default_config,
)
from populus.config.validation import (
    get_validation_errors,
    format_errors,
)
from populus.config.versions import (
    V2,
    V3,
)


NEW_V3_PATHS = {
    'chains.mainnet.contracts.backends.TestContracts',
    'chains.ropsten.contracts.backends.TestContracts',
    'chains.temp.contracts.backends.TestContracts',
    'chains.testrpc.contracts.backends.TestContracts',
    'chains.tester.contracts.backends.TestContracts',
    'contracts.backends.TestContracts',
}


def upgrade_v2_to_v3(v2_config):
    """
    Upgrade a v2 config file to a v3 config file.
    """
    errors = get_validation_errors(v2_config, version=V2)
    if errors:
        raise ValueError(
            "Cannot upgrade invalid config.  Please ensure that your current "
            "configuration file is valid:\n\n{0}".format(
                format_errors(errors),
            )
        )

    v2_default_config = load_default_config(version=V2)
    v3_default_config = load_default_config(version=V3)

    if v2_config == v2_default_config:
        return v3_default_config

    upgraded_v2_config = copy.deepcopy(v2_config)

    for key_path in NEW_V3_PATHS:
        if has_nested_key(upgraded_v2_config, key_path):
            continue
        set_nested_key(
            upgraded_v2_config,
            key_path,
            get_nested_key(v3_default_config, key_path),
        )

    # bump the version
    set_nested_key(upgraded_v2_config, 'version', V3)

    errors = get_validation_errors(upgraded_v2_config, version=V3)
    if errors:
        raise ValueError(
            "Upgraded configuration did not pass validation:\n\n"
            "\n=============Original-Configuration============\n"
            "{0}"
            "\n=============Upgraded-Configuration============\n"
            "{1}"
            "\n=============Validation-Errors============\n"
            "{2}".format(
                pprint.pformat(dict(v2_config)),
                pprint.pformat(dict(upgraded_v2_config)),
                format_errors(errors),
            )
        )

    return upgraded_v2_config
